"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.schemaMapperFactory = void 0;
const x_deep_merge_1 = require("@empathyco/x-deep-merge");
const x_utils_1 = require("@empathyco/x-utils");
const utils_1 = require("../schemas/utils");
/**
 * The 'schemaMapperFactory' function creates a {@link Mapper | mapper function} for a given
 * {@link Schema | schema}.
 *
 * @param schema - The {@link Schema | schema} to apply in the {@link Mapper | mapper function}.
 * @returns A {@link Mapper | mapper function} that applies the given {@link Schema | schema}.
 * @public
 */
function schemaMapperFactory(schema) {
    return function mapper(source, context) {
        return mapSchema(source, schema, context);
    };
}
exports.schemaMapperFactory = schemaMapperFactory;
/**
 * The `mapSchema()` function creates a new object populated with the transformations defined by a
 * {@link Schema} applied to a source object.
 *
 * @param source - The object to apply the transformations to.
 * @param schema - The object that defines the transformations to apply.
 * @param context - The {@link MapperContext | mapper context} to feed the transformations with.
 * @returns A new object with each element being the result of the applied transformation.
 * @internal
 */
function mapSchema(source, schema, context) {
    if (!source) {
        //eslint-disable-next-line no-console
        console.warn('This schema cannot be applied', (0, utils_1.createMutableSchema)(schema));
        return undefined;
    }
    return (0, x_utils_1.reduce)(schema, (target, key, transformer) => {
        if (typeof transformer === 'string' && (0, x_utils_1.isPath)(source, transformer)) {
            target[key] = (0, x_utils_1.getSafePropertyChain)(source, transformer);
        }
        else if ((0, x_utils_1.isFunction)(transformer) && !(0, utils_1.isInternalMethod)(transformer.name)) {
            target[key] = transformer(source, context);
        }
        else if ((0, x_utils_1.isObject)(transformer)) {
            const value = '$subSchema' in transformer
                ? applySubSchemaTransformer(source, transformer, context, schema)
                : mapSchema(source, transformer, context);
            if (value) {
                target[key] = value;
            }
        }
        return target;
    }, {});
}
/**
 * The `applySubSchemaTransformer()` function executes a `mapSchema()` function applying the defined
 * {@link SubSchemaTransformer.$subSchema}.
 *
 * @param source - The object to feed the schema.
 * @param subSchemaTransformer - The {@link SubSchemaTransformer} object with a $path, $subSchema
 * and $context options.
 * @param subSchemaTransformer.$path
 * @param subSchemaTransformer.$subSchema
 * @param rawContext - The {@link MapperContext | mapper context} to feed the mapSchema function.
 * @param subSchemaTransformer.$context
 * @param schema - The {@link Schema} to apply.
 * @returns The result of calling `mapSchema()` with the source, schema and context arguments.
 * @internal
 */
function applySubSchemaTransformer(source, { $subSchema, $path, $context }, rawContext, schema) {
    const subSource = (0, x_utils_1.getSafePropertyChain)(source, $path);
    if (!subSource) {
        return;
    }
    const extendedContext = {};
    if ($context) {
        Object.entries($context).forEach(([key, value]) => {
            if (['requestParameters', 'endpoint', 'mappedValue'].includes(key)) {
                return;
            }
            extendedContext[key] = (0, x_utils_1.isFunction)(value)
                ? value(source)
                : (0, x_utils_1.getSafePropertyChain)(source, value);
        });
    }
    const context = (0, x_deep_merge_1.deepMerge)({}, rawContext, $context, extendedContext);
    let subSchema;
    if ($subSchema === '$self') {
        subSchema = schema;
    }
    else if ((0, x_utils_1.isFunction)($subSchema)) {
        subSchema = $subSchema(source);
    }
    else {
        subSchema = $subSchema;
    }
    return (0, x_utils_1.isArray)(subSource)
        ? subSource.map(item => mapSchema(item, subSchema, context))
        : mapSchema(subSource, subSchema, context);
}
//# sourceMappingURL=schema-mapper.factory.js.map