"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.isInternalMethod = exports.createMutableSchema = void 0;
const x_deep_merge_1 = require("@empathyco/x-deep-merge");
const x_utils_1 = require("@empathyco/x-utils");
/**
 * Collection of internal method names for {@link MutableSchema | mutable schemas}.
 */
const mutableSchemasInternalMethods = ['$replace', '$override', '$extends', 'toString'];
/**
 * Creates a {@link MutableSchema | mutable schema } version of a given {@link Schema | schema}.
 *
 * @param schema - The {@link Schema | schema} to make mutable.
 *
 * @returns A {@link MutableSchema | mutable schema} version of the given {@link Schema | schema}.
 *
 * @public
 */
function createMutableSchema(schema) {
    return {
        ...schema,
        $replace(newSchema) {
            (0, x_utils_1.forEach)(this, key => {
                if (isInternalMethod(key)) {
                    return;
                }
                delete this[key];
            });
            Object.assign(this, newSchema);
            /* We are replacing the schema with a completely new schema , so it makes sense that TS
             complains that the old schema and the new one are not the same. */
            return this;
        },
        $override(newSchema) {
            return (0, x_deep_merge_1.deepMerge)(this, newSchema);
        },
        $extends(newSchema) {
            return (0, x_deep_merge_1.deepMerge)({}, this, newSchema);
        },
        toString(includeInternalMethods = false) {
            return serialize(this, !!includeInternalMethods);
        }
    };
}
exports.createMutableSchema = createMutableSchema;
/**
 * Checks if the given key is a {@link MutableSchema | mutableSchema} method.
 *
 * @param name - The key to check.
 *
 * @returns True if it is a {@link MutableSchema | mutableSchema} method,
 * false otherwise.
 *
 * @public
 */
function isInternalMethod(name) {
    return mutableSchemasInternalMethods.includes(name);
}
exports.isInternalMethod = isInternalMethod;
/**
 * Returns a string representing of the given object.
 *
 * @param data - The object to get the string representation from.
 * @param includeInternalMethods - Flag to include in the string representation
 * the internal methods. Disabled by default.
 * @param deep - The level of indentation.
 * @returns The string representation.
 */
function serialize(data, includeInternalMethods, deep = 0) {
    const indentation = '  '.repeat(deep);
    let output = '';
    (0, x_utils_1.forEach)(data, (key, value) => {
        if ((0, x_utils_1.isObject)(value)) {
            output += `${indentation}${key}: {\n${serialize(value, includeInternalMethods, ++deep)}${indentation}},\n`;
        }
        else if (!(0, x_utils_1.isFunction)(value) || !isInternalMethod(key) || includeInternalMethods) {
            // eslint-disable-next-line @typescript-eslint/restrict-template-expressions
            output += `${indentation}${key}: ${value},\n`;
        }
    });
    return output;
}
//# sourceMappingURL=utils.js.map